OC.L10N.register(
    "mail",
    {
    "Important" : "වැදගත්",
    "Personal" : "පුද්ගලික",
    "Later" : "පසුව",
    "Mail" : "තැපෑල",
    "Save" : "සුරකින්න",
    "Connect" : "සම්බන්ධ කරන්න",
    "Name" : "නම",
    "Password" : "මුර පදය",
    "Signature" : "අත්සන",
    "Go back" : "ආපසු යන්න",
    "Cancel" : "අවලංගු කරන්න",
    "General" : "සමාන්‍යය",
    "Privacy" : "පෞද්ගලිකත්වය",
    "Archive" : "සංරක්ෂණය",
    "Delete" : "Delete",
    "Search" : "සොයන්න",
    "Send" : "යවන්න",
    "About" : "පිළිබඳ",
    "Confirm" : "තහවුරු කරන්න",
    "Choose" : "තෝරන්න",
    "Favorite" : "ප්‍රියතම",
    "Read" : "කියවීම",
    "Back" : "ආපසු",
    "Description" : "විස්තරය",
    "Create" : "සාදන්න",
    "Select" : "තෝරන්න",
    "Comment" : "අදහස",
    "Accept" : "පිළිගන්න",
    "Decline" : "ප්‍රතික්ෂේප",
    "Remove" : "ඉවත් කරන්න",
    "Today" : "අද",
    "No results" : "ප්‍රතිථල නැත",
    "Favorites" : "ප්‍රියතම",
    "Other" : "වෙනත්",
    "Link copied" : "සබැඳිය පිටපත් කළා",
    "Translate" : "පරිවර්තනය",
    "Change password" : "මුරපදය වෙනස් කරන්න",
    "Remove account" : "ගිණුම ඉවත් කරන්න",
    "Folder name" : "බහාලුමේ නම",
    "Mark all as read" : "සියල්ල කියවූ ලෙස සලකුණු කරන්න",
    "Reply" : "පිළිතුර",
    "Retry" : "නැවත උත්සාහ කරන්න",
    "Message" : "පණිවිඩය",
    "Add" : "එකතු කරන්න",
    "Close" : "වසන්න",
    "Body" : "අන්තර්ගතය",
    "Date" : "දිනය",
    "Help" : "උපකාර",
    "Actions" : "ක්‍රියාමාර්ග",
    "Edit" : "සංස්කරණය",
    "User" : "පරිශීලක",
    "Host" : " ධාරකය",
    "Certificate" : "සහතිකය",
    "Shares" : "බෙදාගැනීම්",
    "All" : "සියල්ල",
    "Date:" : "දිනය:",
    "Show more" : "තව පෙන්වන්න"
},
"nplurals=2; plural=(n != 1);");
